import greenfoot.*;

/**
 * CLASS: Twister<br>
 * AUTHOR: danpost (greenfoot.org username)<br>
 * VERSION DATE: January 29, 2014<br>
 * <br>
 * This class will take an image and break it up into smaller and smaller disks (each one a Twister object).
 * These disks (including the initial Twister object) rotate at speeds that increase as the size of the disks decrease
 * producing the 'twister' effect.
 */
public class Twister extends Actor
{
    int speed = 100; // the speed of the initial Twister object created for each twister
    int size; // radius of twister
    Twister sub, sup; // the immediate subordinate (smaller) and superior (larger) Twister objects
    
    /**
     * This constructor creates a twister using an image it creates and random values for radius and point coordinates.
     */
    public Twister()
    {
        size = Greenfoot.getRandomNumber(20)+50; // random radius
        GreenfootImage tiled = new GreenfootImage(size*4, size*4); // create the image that will be the tiled image
        tiled.fill(); // remove all transparency by filling image with black
        // create base image that will be used for tiling
        GreenfootImage base = new GreenfootImage(Greenfoot.getRandomNumber(50)+50, Greenfoot.getRandomNumber(50)+50);
        base.setColor(Color.WHITE);
        base.fillRect(0, 0, base.getWidth()/2, base.getHeight()/2);
        base.fillRect(base.getWidth()/2, base.getHeight()/2, 50, 50);
        // tile the image with the base image
        for (int i=0; i<tiled.getWidth(); i+=base.getWidth())
            for (int j=0; j<tiled.getHeight(); j+=base.getHeight())
                tiled.drawImage(base, i, j);
        GreenfootImage image = new GreenfootImage(size*2, size*2); // create the image for the actor
        int drawX = -Greenfoot.getRandomNumber(base.getWidth()); // determine where base image is to be drawn horizontally
        int drawY = -Greenfoot.getRandomNumber(base.getHeight()); // determine where base image is to be drawn vertically
        image.drawImage(tiled, drawX, drawY); // drawn the base image on future image of actor
        setImage(getRoundedImage(image)); // set the image of the actor to the rounded image
    }
    
    /**
     * This constructor takes a given point as coordinates and a given radius and cuts a circle out of the given
     * image which is tiled.
     *
     * @param base the image to be tiled and the circle to be cut from
     * @param offsetX the x-coordinate of the centerpoint of the circle to be cut from within the tiled image
     * @param offsetY the y-coordinate of the centerpoint of the circle to be cut from within the tiled image
     * @param radius the radius of the circle to be cut from the tiled image at the given coordinates
     */
    public Twister(GreenfootImage base, int offsetX, int offsetY, int radius)
    {
        size = radius; // set the size field
        GreenfootImage tiled = new GreenfootImage(size*4, size*4); // create the image that will be the tiled image
        tiled.fill(); // remove all transparency by filling image with black
        // create the tiled image
        for (int i=0; i<tiled.getWidth(); i+=base.getWidth())
            for (int j=0; j<tiled.getHeight(); j+=base.getHeight())
                tiled.drawImage(base, i, j);
        GreenfootImage image = new GreenfootImage(size*2, size*2); // create the image for the actor
        int drawX = size-offsetX%base.getWidth()-base.getWidth(); // determine where base image is to be drawn horizontally
        int drawY = size-offsetY%base.getHeight()-base.getHeight(); // determine where base image is to be drawn vertically
        image.drawImage(tiled, drawX, drawY); // drawn the base image on future image of actor
        setImage(getRoundedImage(image)); // set the image of the actor to the rounded image
    }
    
    /**
     * This constructor takes a given point as coordinates and a given radius and cuts a circle out of a tiled image
     * whose String filename of the untiled image is given.
     *
     * @param base the String filename of the image to be tiled and the circle to be cut from
     * @param offsetX the x-coordinate of the centerpoint of the circle to be cut from within the tiled image
     * @param offsetY the y-coordinate of the centerpoint of the circle to be cut from within the tiled image
     * @param radius the radius of the circle to be cut from the tiled image at the given coordinates
     */
    public Twister(String base, int offsetX, int offsetY, int radius)
    {
        this(new GreenfootImage(base), offsetX, offsetY, radius);
    }        
    
    /**
     * This constructor is used from within the class for all subsequent Twister objects for a twister after the
     * initial Twister has been created.
     *
     * @param parent the Twister object that instantiated this one
     */
    private Twister(Twister parent)
    {
        sup = parent; // save reference to superior Twister object
        size = sup.size-2; // set size a little smaller than that of parent
        GreenfootImage image = new GreenfootImage(size*2, size*2); // create image
        image.drawImage(sup.getImage(), -2, -2); // draw parent image on this image
        speed = sup.speed+10; // set speed a little faster than that of parent
        setImage(getRoundedImage(image)); // set the image of the Twister object to the rounded image
    }
    
    /**
     * This method, called repeatedly by the system while the scenario is running, rotates the actor and,
     * if it has not created a subordinate Twister object yet and if not limited by size restraints, creates one.
     */
    public void act()
    {
        turn(-speed/10);
        if (sub == null && size*2 > 10)
        {
            sub = new Twister(this);
            getWorld().addObject(sub, getX(), getY());
        }
    }
    
    /**
     * This method removes all Twister objects associated with this twister
     */
    public void remove()
    {
        getWorld().removeObjects(getWorld().getObjectsAt(getX(), getY(), Twister.class));
    }
    
    private GreenfootImage getRoundedImage(GreenfootImage image)
    {
        int half = image.getWidth()/2;
        for (int i=0; i<half*2; i++) for (int j=0; j<half*2; j++)
        {
            if (Math.hypot(i-half, j-half) > half) image.setColorAt(i, j, new Color(0, 0, 0, 0));
            else j = half*2-j;
        }
        return image;
    }
}